;; Copyright (C) 1999-2001 Peter Seibel

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of the
;; License, or (at your option) any later version.

;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
;; USA

(provide 'p4)
(require 'ediff)

(defun p4-find-workspace ()
  (let ((buf (generate-new-buffer "*p4 info*"))
	(workspace nil))
    (shell-command "p4 info" buf)
    (save-excursion
      (set-buffer buf)
      (goto-char (point-min))
      (if (search-forward-regexp "Client root: \\(.*\\)$" (point-max) t)
	  (setq workspace (match-string 1))))
    (kill-buffer buf)
    (delete-other-windows)
    workspace))

(defvar p4-workspace (p4-find-workspace))

(defvar p4-mode-map nil "Local keymap for p4-mode buffers.")
(if p4-mode-map
    nil
  (let ((map (make-keymap)))
    (suppress-keymap map t)
    (define-key map " "  'p4-do-line)
    (define-key map "\C-m" 'p4-do-line)
    (define-key map "q"  (lambda () (interactive) (kill-buffer nil)))
    (setq p4-mode-map map)))

(defvar p4-submit-mode-map nil "Local keymap for p4-submit-mode buffers.")
(if p4-submit-mode-map
    nil
  (let ((map (make-keymap)))
    (define-key map "q"  (lambda () (interactive) (kill-buffer nil)))
    (define-key map [(control x) (?#)] 'p4-finish-submit)
    (setq p4-submit-mode-map map)))

(defvar p4-mode-hook nil
  "Run at the very end of p4-mode.")

(defvar p4-submit-mode-hook nil
  "Run at the very end of p4-submit-mode.")

;; p4 mode is suitable only for specially formatted data.
(put 'p4-mode 'mode-class 'special)

(defun p4-mode ()
  "Mode for browsing the p4 depot."
  (kill-all-local-variables)
  (use-local-map p4-mode-map)
  (setq major-mode 'p4-mode
	mode-name "p4"
	buffer-read-only t
	mode-line-buffer-identification '("%17b"))
  (run-hooks 'p4-mode-hook))

(defun p4-submit-mode ()
  "Mode for preparing submissions."
  (kill-all-local-variables)
  (use-local-map p4-submit-mode-map)
  (setq major-mode 'p4-submit-mode
	mode-name "p4-submit"
	buffer-read-only nil
	mode-line-buffer-identification '("%17b"))
  (run-hooks 'p4-submit-mode-hook))

(defun p4-line-history ()
  "Get the revision history for every line in the current buffer."
  (interactive)
  (p4-shell-to-p4pr))

(defun p4-edit ()
  "p4 edit the file visited by the current buffer."
  (interactive)
  (p4-do-command-on-file "edit"))

(defun p4-add ()
  "p4 add the file visited by the current buffer."
  (interactive)
  (let* ((type (read-from-minibuffer "File type: " "text"))
	 (opts (if type (concat "-t " type) nil)))
    (p4-do-command-on-file "add" opts)))

(defun p4-revert ()
  "p4 revert the file visited by the current buffer and revert the buffer to reflect the actual contents of the file after the revert."
  (interactive)
  (if (yes-or-no-p
       (format "p4 revert %s and buffer? " buffer-file-name))
      (progn
	(p4-do-command-on-file "revert")
	(revert-buffer nil t))))

(defun p4-print (file)
  "p4 print a file to a buffer. Useful for browing old versions."
  (interactive "sFile: ")
  (switch-to-buffer (p4-print-to-buffer file))
  (p4-mode)
  (delete-other-windows))

(defun p4-describe (change)
  "Run p4 describe -s on the given change number."
  (interactive "nChange: ")
  (let* ((string-change (number-to-string change))
	 (buf (generate-new-buffer (concat "*p4 describe: " string-change "*"))))
    (shell-command
     (concat "p4 describe -s " string-change) buf)
    (switch-to-buffer buf)
;    (local-set-key "\C-c\C-c" 'p4-browse-diff)
    (p4-mode)
    (goto-char (point-min))
    (search-forward "... //depot")
    (beginning-of-line)
    (delete-other-windows)))


(defun p4-changes ()
  "Run p4 changes on a pattern and a number of changes."
  (interactive)
  (let* ((pattern (p4-get-pattern))
	 (number (read-from-minibuffer "Number (all): "))
	 (buf (generate-new-buffer "*p4 changes*"))
	 (m (if (not (equal number "")) (concat " -m" number " ") " "))
	 (command (concat "p4 changes" m pattern)))
    (shell-command command buf)
    (switch-to-buffer buf)
    (p4-mode)
;    (local-set-key "\C-c\C-c" 'p4-browse-change)
    (delete-other-windows)))

(defun p4-opened ()
  "Run p4 opened."
  (interactive)
  (let* ((pattern (p4-get-pattern))
	 (buf (generate-new-buffer "*p4 opened*"))
	 (command (concat "p4 opened " pattern)))
    (shell-command command buf)
    (switch-to-buffer buf)
    (p4-mode)
    (delete-other-windows)))

(defun p4-submit ()
  "Run p4 submit"
  (interactive)
  (let* ((buf (generate-new-buffer (concat "*p4 submit*")))
	 (command (concat "p4 change -o")))
    (shell-command command buf)
    (switch-to-buffer buf)
    (p4-submit-mode)
    (delete-other-windows)))

(defun p4-review ()
  "Show a list of changes to be reviewed."
  (interactive)
  (let* ((pattern (p4-get-pattern))
	 (buf (generate-new-buffer (concat "*p4 review*")))
	 (counter (read-from-minibuffer "Counter: " (concat "review-" (p4-user-name))))
	 (counter-value (p4-get-counter-value counter))
	 (command (concat "p4 changes " pattern "...@" counter-value ",")))
    (shell-command command buf)
    (switch-to-buffer buf)
    (reverse-region (point-min) (point-max))
    (p4-mode)
    (delete-other-windows)))

(defun p4-user-name ()
  "Get current p4 user name."
  (interactive)
  (let ((buf (generate-new-buffer "*p4 info*"))
	(result nil))
    (shell-command "p4 info" buf)
    (save-excursion
      (set-buffer buf)
      (goto-char (point-min))
      (search-forward-regexp "User name: \\(.*\\)")
      (setq result (match-string 1)))
    (kill-buffer buf)
    (delete-other-windows)
    result))

(defun p4-get-counter-value (counter-name)
  (interactive "sCounter name: ")
  (let ((buf (generate-new-buffer "*p4 counter*"))
	(command (concat "p4 counter " counter-name))
	(result nil))
    (shell-command command buf)
    (save-excursion
      (set-buffer buf)
      (goto-char (point-min))
      (search-forward-regexp "[0-9]+")
      (setq result (match-string 0))
    (kill-buffer buf)
    (delete-other-windows)
    result)))

(defun p4-set-counter-value (counter-name value)
  (interactive "sCounter name: \nnValue: ")
  (let ((command (concat "p4 counter " counter-name " " (number-to-string value))))
    (shell-command-to-string command)))

(defun p4-finish-submit ()
  (interactive)
  (let* ((buf (generate-new-buffer (concat "*p4 submit<results>*")))
	 (command "p4 submit -i"))
    (shell-command-on-region (point-max) (point-min) command buf nil)
    (switch-to-buffer buf)
    (p4-submit-mode)
    (delete-other-windows)))

(defun p4-get-pattern ()
  "Prompt for a file pattern."
  (interactive)
  (read-from-minibuffer "Pattern: " (p4-get-current-where)))

(defun p4-get-current-where ()
  (let ((buf (generate-new-buffer "*p4 where*"))
	(command (concat "p4 where"))
	(result nil))
    (shell-command command buf)
    (save-excursion
      (set-buffer buf)
      (if (= 1 (count-lines-buffer))
	  (progn
	    (goto-char (point-min))
	    (search-forward-regexp "\\(//depot/.*\.\.\.\\) //")
	    (setq result (match-string 1)))
	(setq result "//depot/..."))
      (kill-buffer buf)
      (delete-other-windows)
      result)))


(defun p4-filelog ()
  "p4 filelog the file visited by the current buffer."
  (interactive)
  (let* ((buf (generate-new-buffer "*p4 filelog*"))
	 (command (concat "p4 filelog -i " (buffer-file-name))))
    (shell-command command buf)
    (switch-to-buffer buf)
    (p4-mode)
    (delete-other-windows)))

(defun p4-do-line ()
  (interactive)
  (cond
   ((string-match "^*p4 opened*" (buffer-name)) (p4-diff-opened))
   ((and (string-match "^*p4 filelog*" (buffer-name))
	 (looking-at "//depot")) (p4-diff2-filelog))
   ((looking-at "... #[0-9]+ change") (p4-browse-filelog))
   ((looking-at ".*Change") (p4-browse-change))
   ((looking-at ".*//depot/") (p4-browse-diff))))

(defun p4-browse-change ()
  "Run p4-describe on the change listed on the current line of
p4-changes output."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (search-forward-regexp "Change \\([0-9]+\\)")
    (p4-describe (string-to-number (match-string 1)))))

(defun p4-browse-diff ()
  "Browse the file on the current line of p4-describe output.
If the file is version 2 or greater p4-diff2 it with the previous version.
Otherwise p4-print it."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (search-forward-regexp "\\(//depot/[^#]+\\)#\\([0-9]+\\)")
    (let ((file (match-string 1))
	  (version (string-to-number (match-string 2))))
      (if (< version 2)
	  (p4-print (concat file "#" (number-to-string version)))
	(p4-diff2
	 (concat file "#" (number-to-string (- version 1)))
	 (concat file "#" (number-to-string version)))))))

(defun p4-browse-filelog ()
  "Browse the file on the current line of the p4-filelog output."
  (interactive)
  (p4-print (concat (p4-filelog-get-basename) (p4-filelog-get-version))))

(defun p4-diff2-filelog ()
  "Diff two versions of the file being filelogged."
  (interactive)
  (let ((filename (p4-diff2-get-basename))
	(old (read-from-minibuffer "Old version: "))
	(new (read-from-minibuffer "New version: ")))
    (p4-diff2 (concat filename "#" old) (concat filename "#" new))))


(defun p4-diff2-get-basename ()
  (save-excursion
    (search-forward-regexp "^\\(//depot/.*\\)$")
    (match-string 1)))

(defun p4-filelog-get-basename ()
  (save-excursion
    (search-backward-regexp "^\\(//depot/.*\\)$")
    (match-string 1)))

(defun p4-filelog-get-version ()
  (save-excursion
    (beginning-of-line)
    (search-forward-regexp "... \\(#[0-9]+\\) change")
    (match-string 1)))

(defun p4-diff (&optional version)
  "p4 diff the file visited by the current buffer with a version in the depot.
If called with a prefix arg, diffs against that version; otherwise diffs
versus the latest version."
  (interactive "P")
  (ediff-buffers
   (current-buffer)
   (p4-make-diff-buffer (p4-print-to-buffer (p4-filename version)))))

(defun p4-diff-opened ()
  "Open the file on the current line."
  (save-excursion
    (beginning-of-line)
    (search-forward-regexp "//depot\\(/[^#]+\\)#\\([0-9]+\\)")
    (let ((file (match-string 1)))
      (find-file (concat p4-workspace file))
      (p4-diff))))

(defun p4-diff2 (file1 file2)
  "Run ediff on any two files in the depot."
  (interactive "sDepot file one: \nsDepot file two: ")
  (save-excursion
    (ediff-buffers
     (p4-make-diff-buffer (p4-print-to-buffer file1))
     (p4-make-diff-buffer (p4-print-to-buffer file2)))))

(defun p4-diff3 (file1 file2 file3)
  "Run ediff on any three files in the depot."
  (interactive "sDepot file one: \nsDepot file two: \nsDepot file three: ")
  (ediff-buffers3
   (p4-make-diff-buffer (p4-print-to-buffer file1))
   (p4-make-diff-buffer (p4-print-to-buffer file2))
   (p4-make-diff-buffer (p4-print-to-buffer file3))))

(defun p4-manual-resolve (file sibling-in-depot original-in-depot)
  "Run ediff on any one actual file and two files from the
depot. Typically used with the actual file being a file in a branch
and the two depot files the latest version of the corresponding file
in the parent branch and the original version of this file in the
branch. This allows you to see changes made in the parent branch and
your branch since the file was created."
  (interactive "sActual file: \nsParent branch: \nsThis branch: ")
  (ediff-buffers3
   (find-file-noselect file)
   (p4-make-diff-buffer (p4-print-to-buffer sibling-in-depot))
   (p4-make-diff-buffer (p4-print-to-buffer original-in-depot))))


(defun p4-diff-resolve ()
  (interactive)
  (p4-diff-resolve-internal
   'p4-make-orig-buffer 'p4-make-theirs-buffer 'p4-make-yours-buffer))

(defun p4-diff-resolve-orig-theirs ()
  (interactive)
  (p4-diff-resolve-internal 'p4-make-orig-buffer 'p4-make-theirs-buffer))

(defun p4-diff-resolve-orig-yours ()
  (interactive)
  (p4-diff-resolve-internal 'p4-make-orig-buffer 'p4-make-yours-buffer))

(defun p4-diff-resolve-theirs-yours ()
  (interactive)
  (p4-diff-resolve-internal 'p4-make-theirs-buffer 'p4-make-yours-buffer))

;; private functions

(defun p4-diff-resolve-internal (fn1 fn2 &optional fn3)
  (search-forward ">>>> ORIGINAL")
  (beginning-of-line)
  (save-excursion
    (if fn3
	(ediff-buffers3
	 (p4-make-diff-buffer (funcall fn1))
	 (p4-make-diff-buffer (funcall fn2))
	 (p4-make-diff-buffer (funcall fn3)))
      (ediff-buffers
       (p4-make-diff-buffer (funcall fn1))
       (p4-make-diff-buffer (funcall fn2))))))

(defun p4-make-orig-buffer ()
  (p4-make-resolve-buffer "ORIGINAL" ">>>> ORIGINAL" "==== THEIRS"))

(defun p4-make-theirs-buffer ()
  (p4-make-resolve-buffer "THEIRS" "==== THEIRS" "==== YOURS"))

(defun p4-make-yours-buffer ()
  (p4-make-resolve-buffer "YOURS" "==== YOURS" "<<<<"))

(defun p4-make-resolve-buffer (label start-tag end-tag)
  (save-excursion
    (let (start
	  end
	  (buf (generate-new-buffer (format "*p4 resolve - %s*" label))))
      (search-forward start-tag)
      (beginning-of-line 2)
      (setq start (point))
      (search-forward end-tag)
      (beginning-of-line)
      (setq end (point))
      (insert-string (buffer-substring start end) buf)
      buf)))

(defun p4-do-command-on-file (command &optional opts)
  (let ((opts (if opts (concat opts " ") "")))
    (shell-command
     (concat "p4 " command " " opts (p4-filename)))
  (setq buffer-read-only (not (file-readable-p buffer-file-name)))))

(defun p4-shell-to-p4pr ()
  (shell-command
   (concat "perl /usr/bin/p4pr.perl " (p4-filename))))

(defun p4-filename (&optional version)
  (let ((base-filename (file-name-nondirectory buffer-file-name)))
    (if version 
	(concat base-filename "#" version)
      base-filename)))

(defun p4-print-to-buffer (filename)
  (let* ((buf (generate-new-buffer (concat filename "<depot>"))))
    (shell-command (concat "p4 print -q " filename) buf)
    buf))

(defvar p4-diff-buffers nil)

(defun p4-make-diff-buffer (buf)
  ; this is redundant but safer than only doing it when
  ; p4-diff-buffers is nil
  (add-hook 'ediff-quit-hook 'p4-cleanup-diff-buffers)
  (setq p4-diff-buffers (cons buf p4-diff-buffers))
  buf)

; This mechanism for cleaning up temporary buffers only works
; assuming that only one instance of ediff is run at a time. Which
; seems safe enough for the moment.
(defun p4-cleanup-diff-buffers ()
  (mapcar (lambda (b) (kill-buffer b)) p4-diff-buffers)
  (setq p4-diff-buffers nil)
  (remove-hook 'ediff-quit-hook 'p4-cleanup-diff-buffers))
